import singer
import tap_framework

import tap_chargebee.client
import tap_chargebee.streams

LOGGER = singer.get_logger()


class ChargebeeRunner(tap_framework.Runner):
    pass


@singer.utils.handle_top_exception(LOGGER)
def main():
    args = singer.utils.parse_args(
        required_config_keys=['api_key', 'start_date', 'site'])

    client = tap_chargebee.client.ChargebeeClient(args.config)

    runner = ChargebeeRunner(
        args, client, get_available_streams(args, client)
    )

    try:
        # Verify start date format
        singer.utils.strptime(args.config.get("start_date"))
    except ValueError:
        raise ValueError("start_date must be in 'YYYY-mm-ddTHH:MM:SSZ' format") from None

    if args.discover:
        runner.do_discover()
    else:
        runner.do_sync()


if __name__ == '__main__':
    main()


def get_available_streams(self, cb_client):
    site_name = self.config.get('site')
    LOGGER.info("Site Name {}".format(site_name))
    configuration_url = 'https://{}.chargebee.com/api/v2/configurations'.format(site_name)
    response = cb_client.make_request(
        url=configuration_url,
        method='GET')
    site_configurations = response['configurations']
    LOGGER.info("Configurations API response {}".format(response))
    product_catalog_version = next(iter(config['product_catalog_version'] for config in site_configurations if
                                        config['domain'].lower() == site_name.lower()),
                                   None)
    if product_catalog_version == 'v2':
        available_streams = tap_chargebee.streams.ITEM_MODEL_AVAILABLE_STREAMS
        self.config['item_model'] = True
        LOGGER.info('Item Model')
    elif product_catalog_version == 'v1':
        available_streams = tap_chargebee.streams.PLAN_MODEL_AVAILABLE_STREAMS
        self.config['item_model'] = False
        LOGGER.info('Plan Model')
    else:
        LOGGER.error("Incorrect Product Catalog version {}".format(product_catalog_version))
        raise RuntimeError("Incorrect Product Catalog version")
    return available_streams



import datetime
import json
import singer

LOGGER = singer.get_logger()


def get_last_record_value_for_table(state, table, field):
    last_value = state.get('bookmarks', {}) \
                      .get(table, {}) \
                      .get(field)

    if last_value is None:
        return None

    return last_value


def incorporate(state, table, key, value, force=False):
    if value is None:
        return state

    if isinstance(value, datetime.datetime):
        value = value.strftime('%Y-%m-%dT%H:%M:%SZ')

    if state is None:
        new_state = {}
    else:
        new_state = state.copy()

    if 'bookmarks' not in new_state:
        new_state['bookmarks'] = {}

    if table not in new_state['bookmarks']:
        new_state['bookmarks'][table] = {}

    if(new_state['bookmarks'].get(table, {}).get(key) is None or
       new_state['bookmarks'].get(table, {}).get(key) < value or
       force):
        new_state['bookmarks'][table][key] = value

    return new_state


def save_state(state):
    if not state:
        return

    LOGGER.info('Updating state.')

    singer.write_state(state)


def load_state(filename):
    if filename is None:
        return {}

    try:
        with open(filename) as handle:
            return json.load(handle)
    except:
        LOGGER.fatal("Failed to decode state file. Is it valid json?")
        raise RuntimeError



import singer
import time
import json
import os

import pytz

from .util import Util
from dateutil.parser import parse
from tap_framework.streams import BaseStream
from tap_framework.schemas import load_schema_by_name
from tap_framework.config import get_config_start_date
from tap_chargebee.state import get_last_record_value_for_table, incorporate, \
    save_state

LOGGER = singer.get_logger()


class BaseChargebeeStream(BaseStream):

    def write_schema(self):
        singer.write_schema(
            self.catalog.stream,
            self.catalog.schema.to_dict(),
            key_properties=self.KEY_PROPERTIES,
            bookmark_properties=self.BOOKMARK_PROPERTIES)

    def get_abs_path(self, path):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)

    def load_shared_schema_refs(self):
        """Select folder to create a reference dict."""
        shared_schema_refs = {}
        schema_folders = ["common"]
        if self.config['item_model']:
            # Chosen streams of product catalog v2
            schema_folders.append("item_model")
        else:
            # Chosen streams of product catalog v1
            schema_folders.append("plan_model")
        for schema_folder in schema_folders:
            shared_schema_refs.update(self.load_shared_schema_ref(schema_folder))
        return shared_schema_refs

    def load_shared_schema_ref(self,folder_name):
        """Create a reference dict of all streams."""
        shared_schemas_path = self.get_abs_path('../schemas/'+folder_name)

        shared_file_names = [f for f in os.listdir(shared_schemas_path)
                            if os.path.isfile(os.path.join(shared_schemas_path, f))]

        shared_schema_refs = {}
        for shared_file in shared_file_names:
            # Excluded event stream as it is not used as a reference in any other stream
            if shared_file == "events.json":
                continue
            with open(os.path.join(shared_schemas_path, shared_file)) as data_file:
                shared_schema_refs[shared_file] = json.load(data_file)

        return shared_schema_refs

    def generate_catalog(self):
        schema = self.get_schema()
        mdata = singer.metadata.new()

        metadata = {

            "forced-replication-method": self.REPLICATION_METHOD,
            "valid-replication-keys": self.VALID_REPLICATION_KEYS,
            "inclusion": self.INCLUSION,
            #"selected-by-default": self.SELECTED_BY_DEFAULT,
            "table-key-properties": self.KEY_PROPERTIES
        }

        for k, v in metadata.items():
            mdata = singer.metadata.write(
                mdata,
                (),
                k,
                v
            )

        for field_name, field_schema in schema.get('properties').items():
            inclusion = 'available'

            if field_name in self.KEY_PROPERTIES or field_name in self.BOOKMARK_PROPERTIES:
                inclusion = 'automatic'

            mdata = singer.metadata.write(
                mdata,
                ('properties', field_name),
                'inclusion',
                inclusion
            )

        refs = self.load_shared_schema_refs()

        return [{
            'tap_stream_id': self.TABLE,
            'stream': self.TABLE,
            'schema': singer.resolve_schema_references(schema, refs),
            'metadata': singer.metadata.to_list(mdata)
        }]

    def appendCustomFields(self, record):
        listOfCustomFieldObj = ['addon', 'plan', 'subscription', 'customer']
        custom_fields = {}
        event_custom_fields = {}
        if self.ENTITY == 'event':
            content = record['content']
            words = record['event_type'].split("_")
            sl = slice(len(words) - 1)
            content_obj = "_".join(words[sl])     
            
            if content_obj in listOfCustomFieldObj:
                for k in record['content'][content_obj].keys():
                    if "cf_" in k:
                        event_custom_fields[k] = record['content'][content_obj][k]
                record['content'][content_obj]['custom_fields'] = json.dumps(event_custom_fields)    
        

        for key in record.keys():
            if "cf_" in key:
                custom_fields[key] = record[key]
        if custom_fields:
            record['custom_fields'] = json.dumps(custom_fields)
        return record

    # This overrides the transform_record method in the Fistown Analytics tap-framework package
    def transform_record(self, record):
        with singer.Transformer(integer_datetime_fmt="unix-seconds-integer-datetime-parsing") as tx:
            metadata = {}
            
            record = self.appendCustomFields(record)
                
            if self.catalog.metadata is not None:
                metadata = singer.metadata.to_map(self.catalog.metadata)

            return tx.transform(
                record,
                self.catalog.schema.to_dict(),
                metadata)

    def get_stream_data(self, data):
        entity = self.ENTITY
        return [self.transform_record(item.get(entity)) for item in data]

    def sync_data(self):
        table = self.TABLE
        api_method = self.API_METHOD
        done = False

        # Attempt to get the bookmark date from the state file (if one exists and is supplied).
        LOGGER.info('Attempting to get the most recent bookmark_date for entity {}.'.format(self.ENTITY))
        bookmark_date = get_last_record_value_for_table(self.state, table, 'bookmark_date')

        # If there is no bookmark date, fall back to using the start date from the config file.
        if bookmark_date is None:
            LOGGER.info('Could not locate bookmark_date from STATE file. Falling back to start_date from config.json instead.')
            bookmark_date = get_config_start_date(self.config)
        else:
            bookmark_date = parse(bookmark_date)

        # Convert bookmarked start date to POSIX.
        bookmark_date_posix = int(bookmark_date.timestamp())
        
        # Create params for filtering
        if self.ENTITY == 'event':
            params = {"occurred_at[after]": bookmark_date_posix}
            bookmark_key = 'occurred_at'
        elif self.ENTITY == 'promotional_credit':
            params = {"created_at[after]": bookmark_date_posix}
            bookmark_key = 'created_at'
        else:
            params = {"updated_at[after]": bookmark_date_posix}
            bookmark_key = 'updated_at'

        # Add sort_by[asc] to prevent data overwrite by oldest deleted records
        if self.SORT_BY is not None:
            params['sort_by[asc]'] = self.SORT_BY

        LOGGER.info("Querying {} starting at {}".format(table, bookmark_date))

        while not done:
            max_date = bookmark_date

            response = self.client.make_request(
                url=self.get_url(),
                method=api_method,
                params=params)

            if 'api_error_code' in response.keys():
                if response['api_error_code'] == 'configuration_incompatible':
                    LOGGER.error('{} is not configured'.format(response['error_code']))
                    break

            records = response.get('list')
            
            to_write = self.get_stream_data(records)
            
            if self.ENTITY == 'event':
                for event in to_write:
                    if event["event_type"] == 'plan_deleted':
                        Util.plans.append(event['content']['plan'])
                    elif event['event_type'] == 'addon_deleted':
                        Util.addons.append(event['content']['addon'])
                    elif event['event_type'] == 'coupon_deleted':
                        Util.coupons.append(event['content']['coupon'])
            if self.ENTITY == 'plan':
                for plan in Util.plans:
                    to_write.append(plan)
            if self.ENTITY == 'addon':
                for addon in Util.addons:
                    to_write.append(addon)
            if self.ENTITY == 'coupon':
                for coupon in Util.coupons:
                    to_write.append(coupon) 

            
            with singer.metrics.record_counter(endpoint=table) as ctr:
                singer.write_records(table, to_write)

                ctr.increment(amount=len(to_write))
                
                for item in to_write:
                    #if item.get(bookmark_key) is not None:
                    max_date = max(
                        max_date,
                        parse(item.get(bookmark_key))
                    )

            self.state = incorporate(
                self.state, table, 'bookmark_date', max_date)

            if not response.get('next_offset'):
                LOGGER.info("Final offset reached. Ending sync.")
                done = True
            else:
                LOGGER.info("Advancing by one offset.")
                params['offset'] = response.get('next_offset')
                bookmark_date = max_date

        save_state(self.state)

    def get_schema(self):
        return self.load_schema_by_name(self.SCHEMA)



from tap_chargebee.streams.base import BaseChargebeeStream


class VirtualBankAccountsStream(BaseChargebeeStream):
    TABLE = 'virtual_bank_accounts'
    ENTITY = 'virtual_bank_account'
    REPLICATION_METHOD = 'INCREMENTAL'
    REPLICATION_KEY = 'updated_at'
    KEY_PROPERTIES = ['id']
    BOOKMARK_PROPERTIES = ['updated_at']
    SELECTED_BY_DEFAULT = True
    VALID_REPLICATION_KEYS = ['updated_at']
    INCLUSION = 'available'
    API_METHOD = 'GET'
    SCHEMA = 'common/virtual_bank_accounts'
    SORT_BY = None

    def get_url(self):
        return 'https://{}.chargebee.com/api/v2/virtual_bank_accounts'.format(self.config.get('site'))



from .addons import AddonsStream
from .coupons import CouponsStream
from .credit_notes import CreditNotesStream
from .customers import CustomersStream
from .events import EventsStream
from .invoices import InvoicesStream
from .item_families import ItemFamiliesStream
from .item_prices import ItemPricesStream
from .items import ItemsStream
from .payment_sources import PaymentSourcesStream
from .plans import PlansStream
from .subscriptions import SubscriptionsStream
from .transactions import TransactionsStream
from .virtual_bank_accounts import VirtualBankAccountsStream
from .credit_notes import CreditNotesStream
from .gifts import GiftsStream
from .orders import OrdersStream
from .promotional_credits import PromotionalCreditsStream

COMMON_AVAILABLE_STREAMS = [
    EventsStream,
    CouponsStream,
    CreditNotesStream,
    CustomersStream,
    GiftsStream,
    InvoicesStream,
    OrdersStream,
    PaymentSourcesStream,
    PromotionalCreditsStream,
    SubscriptionsStream,
    TransactionsStream,
    VirtualBankAccountsStream
]

PLAN_MODEL_AVAILABLE_STREAMS = COMMON_AVAILABLE_STREAMS + [
    AddonsStream,
    PlansStream
]

ITEM_MODEL_AVAILABLE_STREAMS = COMMON_AVAILABLE_STREAMS + [
    ItemsStream,
    ItemPricesStream,
    ItemFamiliesStream
]



import backoff
import time
import requests
import singer
import json
import simplejson

from singer import utils
from tap_framework.client import BaseClient


LOGGER = singer.get_logger()


class Server4xxError(Exception):
    pass


class Server429Error(Exception):
    pass


class ChargebeeClient(BaseClient):

    def __init__(self, config, api_result_limit=100, include_deleted=True):
        super().__init__(config)

        self.api_result_limit = api_result_limit
        self.include_deleted = include_deleted
        self.user_agent = self.config.get('user_agent')

    def get_headers(self):
        headers = {}

        if self.config.get('user_agent'):
            headers['User-Agent'] = self.config.get('user_agent')

        return headers

    def get_params(self, params):

        if params is None:
            params = {}

        params['limit'] = self.api_result_limit
        params['include_deleted'] = self.include_deleted

        return params

    @backoff.on_exception(backoff.expo,
                          (Server4xxError, Server429Error),
                          max_tries=5,
                          factor=3)
    @utils.ratelimit(100, 60)
    def make_request(self, url, method, params=None, body=None):

        if params is None:
            params = {}

        LOGGER.info("Making {} request to {}".format(method, url))

        response = requests.request(
            method,
            url,
            auth=(self.config.get("api_key"), ''),
            headers=self.get_headers(),
            params=self.get_params(params),
            json=body)

        try:
            response_json = response.json()
        except simplejson.scanner.JSONDecodeError:
            # Formatted error message for json decoder error
            response_json = {
                "message": "Did not get response from the server due to an unknown error.",
                "http_status_code": response.status_code
            }

        if response.status_code == 429:
            raise Server429Error()

        if response.status_code >= 400:
            raise Server4xxError(response_json)

        return response_json



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='tap-chargebee',
      version='1.2.2',
      description='Singer.io tap for extracting data from the Chargebee API',
      author='dwallace@envoy.com',
      classifiers=['Programming Language :: Python :: 3 :: Only'],
      py_modules=['tap_chargebee'],
      install_requires=[
          'tap-framework==0.1.1'
      ],
      entry_points='''
          [console_scripts]
          tap-chargebee=tap_chargebee:main
      ''',
      packages=find_packages(),
      package_data={
          'tap_chargebee': [
              'schemas/common/*.json',
              'schemas/item_model/*.json',
              'schemas/plan_model/*.json'
          ]
      },
      include_package_data=True)
